@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2009-2019 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2009-2021 -- leonerd@leonerd.org.uk
 
 package IO::Async::LoopTests;
 
@@ -15,6 +15,7 @@ our @EXPORT = qw(
 
 use Test::More;
 use Test::Fatal;
+use Test::Metrics::Any;
 use Test::Refcount;
 
 use IO::Async::Test qw();
@@ -27,7 +28,7 @@ use POSIX qw( SIGTERM );
 use Socket qw( sockaddr_family AF_UNIX );
 use Time::HiRes qw( time );
 
-our $VERSION = '0.75';
+our $VERSION = '0.78';
 
 # Abstract Units of Time
 use constant AUT => $ENV{TEST_QUICK_TIMERS} ? 0.1 : 1;
@@ -89,11 +90,6 @@ sub run_tests
 {
    my ( $testclass, @tests ) = @_;
 
-   my $count = 0;
-   $count += __PACKAGE__->can( "count_tests_$_" )->() + 4 for @tests;
-
-   plan tests => $count;
-
    ( my $file = "$testclass.pm" ) =~ s{::}{/}g;
 
    eval { require $file };
@@ -118,6 +114,8 @@ sub run_tests
 
       is_oneref( $loop, '$loop has refcount 1 finally' );
    }
+
+   done_testing;
 }
 
 sub wait_for(&)
@@ -160,7 +158,6 @@ Tests the Loop's ability to watch filehandles for IO readiness
 
 =cut
 
-use constant count_tests_io => 18;
 sub run_tests_io
 {
    {
@@ -429,61 +426,73 @@ Tests the Loop's ability to handle timer events
 
 =cut
 
-use constant count_tests_timer => 21;
 sub run_tests_timer
 {
-   my $done = 0;
    # New watch/unwatch API
 
    cmp_ok( abs( $loop->time - time ), "<", 0.1, '$loop->time gives the current time' );
 
-   $loop->watch_time( after => 2 * AUT, code => sub { $done = 1; } );
+   # ->watch_time after
+   {
+      my $done;
+      $loop->watch_time( after => 2 * AUT, code => sub { $done = 1; } );
 
-   is_oneref( $loop, '$loop has refcount 1 after watch_time' );
+      is_oneref( $loop, '$loop has refcount 1 after watch_time' );
 
-   time_between {
-      my $now = time;
-      $loop->loop_once( 5 * AUT );
+      time_between {
+         my $now = time;
+         $loop->loop_once( 5 * AUT );
 
-      # poll might have returned just a little early, such that the TimerQueue
-      # doesn't think anything is ready yet. We need to handle that case.
-      while( !$done ) {
-         die "It should have been ready by now" if( time - $now > 5 * AUT );
-         $loop->loop_once( 0.1 * AUT );
-      }
-   } 1.5, 2.5, 'loop_once(5) while waiting for watch_time after';
+         # poll might have returned just a little early, such that the TimerQueue
+         # doesn't think anything is ready yet. We need to handle that case.
+         while( !$done ) {
+            die "It should have been ready by now" if( time - $now > 5 * AUT );
+            $loop->loop_once( 0.1 * AUT );
+         }
+      } 1.5, 2.5, 'loop_once(5) while waiting for watch_time after';
+   }
 
-   $loop->watch_time( at => time + 2 * AUT, code => sub { $done = 2; } );
+   # ->watch_time at
+   {
+      my $done;
+      $loop->watch_time( at => time + 2 * AUT, code => sub { $done = 1; } );
 
-   time_between {
-      my $now = time;
-      $loop->loop_once( 5 * AUT );
+      time_between {
+         my $now = time;
+         $loop->loop_once( 5 * AUT );
 
-      # poll might have returned just a little early, such that the TimerQueue
-      # doesn't think anything is ready yet. We need to handle that case.
-      while( !$done ) {
-         die "It should have been ready by now" if( time - $now > 5 * AUT );
-         $loop->loop_once( 0.1 * AUT );
-      }
-   } 1.5, 2.5, 'loop_once(5) while waiting for watch_time at';
+         # poll might have returned just a little early, such that the TimerQueue
+         # doesn't think anything is ready yet. We need to handle that case.
+         while( !$done ) {
+            die "It should have been ready by now" if( time - $now > 5 * AUT );
+            $loop->loop_once( 0.1 * AUT );
+         }
+      } 1.5, 2.5, 'loop_once(5) while waiting for watch_time at';
+   }
 
-   my $cancelled_fired = 0;
-   my $id = $loop->watch_time( after => 1 * AUT, code => sub { $cancelled_fired = 1 } );
-   $loop->unwatch_time( $id );
-   undef $id;
+   # cancelled timer
+   {
+      my $cancelled_fired = 0;
+      my $id = $loop->watch_time( after => 1 * AUT, code => sub { $cancelled_fired = 1 } );
+      $loop->unwatch_time( $id );
+      undef $id;
 
-   $loop->loop_once( 2 * AUT );
+      $loop->loop_once( 2 * AUT );
 
-   ok( !$cancelled_fired, 'unwatched watch_time does not fire' );
+      ok( !$cancelled_fired, 'unwatched watch_time does not fire' );
+   }
 
-   $loop->watch_time( after => -1, code => sub { $done = 1 } );
+   # ->watch_after negative time
+   {
+      my $done;
+      $loop->watch_time( after => -1, code => sub { $done = 1 } );
 
-   $done = 0;
-
-   time_between {
-      $loop->loop_once while !$done;
-   } 0, 0.1, 'loop_once while waiting for negative interval timer';
+      time_between {
+         $loop->loop_once while !$done;
+      } 0, 0.1, 'loop_once while waiting for negative interval timer';
+   }
 
+   # self-cancellation
    {
       my $done;
 
@@ -501,25 +510,6 @@ sub run_tests_timer
       is( $done, 1, 'Other timers still fire after self-cancelling one' );
    }
 
-   # Legacy enqueue/requeue/cancel API
-   $done = 0;
-
-   $loop->enqueue_timer( delay => 2 * AUT, code => sub { $done = 1; } );
-
-   is_oneref( $loop, '$loop has refcount 1 after enqueue_timer' );
-
-   time_between {
-      my $now = time;
-      $loop->loop_once( 5 * AUT );
-
-      # poll might have returned just a little early, such that the TimerQueue
-      # doesn't think anything is ready yet. We need to handle that case.
-      while( !$done ) {
-         die "It should have been ready by now" if( time - $now > 5 * AUT );
-         $loop->loop_once( 0.1 * AUT );
-      }
-   } 1.5, 2.5, 'loop_once(5) while waiting for timer';
-
    SKIP: {
       skip "Unable to handle sub-second timers accurately", 3 unless $loop->_CAN_SUBSECOND_ACCURATELY;
 
@@ -529,7 +519,7 @@ sub run_tests_timer
          my $count = 0;
          my $start = time;
 
-         $loop->enqueue_timer( delay => $delay, code => sub { $done++ } );
+         $loop->watch_timer( delay => $delay, code => sub { $done++ } );
 
          while( !$done ) {
             $loop->loop_once( 1 );
@@ -540,38 +530,6 @@ sub run_tests_timer
          is( $count, 1, "One ->loop_once(1) sufficient for a single $delay second timer" );
       }
    }
-
-   $cancelled_fired = 0;
-   $id = $loop->enqueue_timer( delay => 1 * AUT, code => sub { $cancelled_fired = 1 } );
-   $loop->cancel_timer( $id );
-   undef $id;
-
-   $loop->loop_once( 2 * AUT );
-
-   ok( !$cancelled_fired, 'cancelled timer does not fire' );
-
-   $id = $loop->enqueue_timer( delay => 1 * AUT, code => sub { $done = 2; } );
-   $id = $loop->requeue_timer( $id, delay => 2 * AUT );
-
-   $done = 0;
-
-   time_between {
-      $loop->loop_once( 1 * AUT );
-
-      is( $done, 0, '$done still 0 so far' );
-
-      my $now = time;
-      $loop->loop_once( 5 * AUT );
-
-      # poll might have returned just a little early, such that the TimerQueue
-      # doesn't think anything is ready yet. We need to handle that case.
-      while( !$done ) {
-         die "It should have been ready by now" if( time - $now > 5 * AUT );
-         $loop->loop_once( 0.1 * AUT );
-      }
-   } 1.5, 2.5, 'requeued timer of delay 2';
-
-   is( $done, 2, '$done is 2 after requeued timer' );
 }
 
 =head2 signal
@@ -580,7 +538,6 @@ Tests the Loop's ability to watch POSIX signals
 
 =cut
 
-use constant count_tests_signal => 15;
 sub run_tests_signal
 {
    unless( IO::Async::OS->HAVE_SIGNALS ) {
@@ -669,7 +626,6 @@ Tests the Loop's support for idle handlers
 
 =cut
 
-use constant count_tests_idle => 11;
 sub run_tests_idle
 {
    my $called = 0;
@@ -702,6 +658,10 @@ sub run_tests_idle
 
    $loop->unwatch_idle( $id );
 
+   # Some loop types (e.g. UV) need to clear a pending queue first and thus the
+   # first loop_once will take zero time
+   $loop->loop_once( 0 );
+
    time_between { $loop->loop_once( 1 * AUT ) } 0.5, 1.5, 'loop_once(1) with unwatched deferral';
 
    is( $called, 2, 'unwatched deferral not called' );
@@ -722,10 +682,12 @@ sub run_tests_idle
    is( $called, 4, '$loop->later shortcut works' );
 }
 
-=head2 child
+=head2 process
 
 Tests the Loop's support for watching child processes by PID
 
+(Previously called C<child>)
+
 =cut
 
 sub run_in_child(&)
@@ -738,8 +700,7 @@ sub run_in_child(&)
    die "Fell out of run_in_child!\n";
 }
 
-use constant count_tests_child => 7;
-sub run_tests_child
+sub run_tests_process
 {
    my $kid = run_in_child {
       exit( 3 );
@@ -747,9 +708,9 @@ sub run_tests_child
 
    my $exitcode;
 
-   $loop->watch_child( $kid => sub { ( undef, $exitcode ) = @_; } );
+   $loop->watch_process( $kid => sub { ( undef, $exitcode ) = @_; } );
 
-   is_oneref( $loop, '$loop has refcount 1 after watch_child' );
+   is_oneref( $loop, '$loop has refcount 1 after watch_process' );
    ok( !defined $exitcode, '$exitcode not defined before ->loop_once' );
 
    undef $exitcode;
@@ -772,7 +733,7 @@ sub run_tests_child
          exit( 0 );
       };
 
-      $loop->watch_child( $kid => sub { ( undef, $exitcode ) = @_; } );
+      $loop->watch_process( $kid => sub { ( undef, $exitcode ) = @_; } );
 
       kill SIGTERM, $kid;
 
@@ -782,17 +743,31 @@ sub run_tests_child
       is( ($exitcode & 0x7f), SIGTERM, 'WTERMSIG($exitcode) after SIGTERM' );
    }
 
-   my %kids;
+   SKIP: {
+      my %kids;
 
-   $loop->watch_child( 0 => sub { my ( $kid ) = @_; delete $kids{$kid} } );
+      $loop->_CAN_WATCH_ALL_PIDS or skip "Loop cannot watch_process for all PIDs", 2;
 
-   %kids = map { run_in_child { exit 0 } => 1 } 1 .. 3;
+      $loop->watch_process( 0 => sub { my ( $kid ) = @_; delete $kids{$kid} } );
 
-   is( scalar keys %kids, 3, 'Waiting for 3 child processes' );
+      %kids = map { run_in_child { exit 0 } => 1 } 1 .. 3;
 
-   wait_for { !keys %kids };
-   ok( !keys %kids, 'All child processes reclaimed' );
+      is( scalar keys %kids, 3, 'Waiting for 3 child processes' );
+
+      wait_for { !keys %kids };
+      ok( !keys %kids, 'All child processes reclaimed' );
+   }
+
+   # Legacy API name
+   $kid = run_in_child { exit 2 };
+
+   undef $exitcode;
+   $loop->watch_child( $kid => sub { ( undef, $exitcode ) = @_; } );
+   wait_for { defined $exitcode };
+
+   is( ($exitcode >> 8), 2, '$exitcode after child exit from legacy ->watch_child' );
 }
+*run_tests_child = \&run_tests_process; # old name
 
 =head2 control
 
@@ -801,7 +776,6 @@ behave correctly
 
 =cut
 
-use constant count_tests_control => 9;
 sub run_tests_control
 {
    time_between { $loop->loop_once( 0 ) } 0, 0.1, 'loop_once(0) when idle';
@@ -856,6 +830,39 @@ sub run_tests_control
    ok( 1, '$loop->loop_forever interruptable by ->loop_stop' );
 }
 
+=head2 metrics
+
+Tests that metrics are generated appropriately using L<Metrics::Any>.
+
+=cut
+
+sub run_tests_metrics
+{
+   my $loopclass = ref $loop;
+
+   return unless $IO::Async::Metrics::METRICS;
+
+   # We should already at least have the loop-type metric
+   is_metrics(
+      {
+         "io_async_loops class:$loopclass" => 1,
+      },
+      'Constructing the loop creates a loop type metric'
+   );
+
+   # The very first call won't create timing metrics because it isn't armed yet.
+   $loop->loop_once( 0 );
+
+   is_metrics_from(
+      sub { $loop->loop_once( 0.1 ) },
+      {
+         io_async_processing_count => 1,
+         io_async_processing_total => Test::Metrics::Any::positive,
+      },
+      'loop_once(0) creates timing metrics'
+   );
+}
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
